"""
MIT License

Copyright (c) 2022 Ramon Hagenaars

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import re
from collections import Counter, defaultdict
from difflib import get_close_matches
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Mapping,
    Union,
)

import numpy as np

from nptyping.error import InvalidStructureError
from nptyping.typing_ import StructureExpression

if TYPE_CHECKING:
    from nptyping.structure import Structure  # pragma: no cover


def validate_structure_expression(
    structure_expression: Union[StructureExpression, Any]
) -> None:
    """
    Validate the given structure_expression and raise an InvalidStructureError
    if it is deemed invalid.
    :param structure_expression: the structure expression in question.
    :return: None.
    """
    if structure_expression is not Any:
        if not re.match(_REGEX_STRUCTURE_EXPRESSION, structure_expression):
            raise InvalidStructureError(
                f"'{structure_expression}' is not a valid structure expression."
            )
        _validate_structure_expression_contains_no_multiple_field_names(
            structure_expression
        )


def check_structure(
    structured_dtype: np.dtype,  # type: ignore[type-arg]
    target: "Structure",
    type_per_name: Dict[str, type],
) -> bool:
    """
    Check the given structured_dtype against the given target Structure and
    return whether it corresponds (True) or not (False). The given dictionary
    contains the vocabulary context for the check.
    :param structured_dtype: the dtype in question.
    :param target: the target Structure that is checked against.
    :param type_per_name: a dict that holds the types by their names as they
    occur in a structure expression.
    :return: True if the given dtype is valid with the given target.
    """
    fields: Mapping[str, Any] = structured_dtype.fields or {}  # type: ignore[assignment]
    for name, dtype_tuple in fields.items():
        dtype = dtype_tuple[0]
        target_type_name = target.get_type(name)
        check_type_name(target_type_name, type_per_name)
        target_type = type_per_name[target_type_name]
        if not issubclass(dtype.type, target_type):
            return False
    return True


def check_type_names(structure: "Structure", type_per_name: Dict[str, type]) -> None:
    """
    Check the given structure for any invalid type names in the given context
    of type_per_name. Raises an InvalidStructureError if a type name is
    invalid.
    :param structure: the Structure that is checked.
    :param type_per_name: the context that determines which type names are valid.
    :return: None.
    """
    for type_ in structure.get_types():
        check_type_name(type_, type_per_name)


def check_type_name(type_name: str, type_per_name: Dict[str, type]) -> None:
    """
    Check if the given type_name is in type_per_name and raise a meaningful
    error if not.
    :param type_name: the key that is checked to be in type_per_name.
    :param type_per_name: a dict that is looked in for type_name.
    :return: None.
    """
    if type_name not in type_per_name:
        close_matches = get_close_matches(
            type_name, type_per_name.keys(), 3, cutoff=0.4
        )
        close_matches_str = ", ".join(f"'{match}'" for match in close_matches)
        extra_help = ""
        if len(close_matches) > 1:
            extra_help = f" Did you mean one of {close_matches_str}?"
        elif close_matches:
            extra_help = f" Did you mean {close_matches_str}?"
        raise InvalidStructureError(  # pylint: disable=raise-missing-from
            f"Type '{type_name}' is not valid in this context.{extra_help}"
        )


def normalize_structure_expression(
    structure_expression: StructureExpression,
) -> StructureExpression:
    """
    Normalize the given structure expression, e.g. by removing whitespaces,
    making similar expressions look the same.
    :param structure_expression: the structure expression that is to be normalized.
    :return: a normalized structure expression.
    """
    structure_expression = re.sub(r"\s*", "", structure_expression)
    type_to_names_dict = _create_type_to_names_dict(structure_expression)
    normalized_structure_expression = _type_to_names_dict_to_str(type_to_names_dict)
    return normalized_structure_expression.replace(",", ", ")


def create_name_to_type_dict(
    structure_expression: StructureExpression,
) -> Dict[str, str]:
    """
    Create a dict with a name as key and a type (str) as value from the given
    structure expression. Structure["x: Int, y: Float"] would yield
    {"x: "Int", "y": "Float"}.
    :param structure_expression: the structure expression from which the dict
    is extracted.
    :return: a dict with names and their types, both as strings.
    """
    type_to_names_dict = _create_type_to_names_dict(structure_expression)
    return {
        name.strip(): type_.strip()
        for type_, names in type_to_names_dict.items()
        for name in names
    }


def _validate_structure_expression_contains_no_multiple_field_names(
    structure_expression: StructureExpression,
) -> None:
    # Validate that there are not multiple occurrences of the same field names.
    matches = re.findall(_REGEX_FIELD, re.sub(r"\s*", "", structure_expression))
    field_name_combinations = [match[0].split(":")[0] for match in matches]
    field_names: List[str] = []
    for field_name_combination in field_name_combinations:
        field_name_combination_match = re.match(
            _REGEX_FIELD_NAMES_COMBINATION, field_name_combination
        )
        if field_name_combination_match:
            field_names += field_name_combination_match.group(2).split(_SEPARATOR)
        else:
            field_names.append(field_name_combination)
    field_name_counter = Counter(field_names)
    field_names_occurring_multiple_times = [
        field_name for field_name, amount in field_name_counter.items() if amount > 1
    ]
    if field_names_occurring_multiple_times:
        # If there are multiple, just raise about the first. Otherwise the
        # error message gets bloated.
        field_name_under_fire = field_names_occurring_multiple_times[0]
        raise InvalidStructureError(
            f"Field names may occur only once in a structure expression."
            f" Field name '{field_name_under_fire}' occurs"
            f" {field_name_counter[field_name_under_fire]} times in"
            f" '{structure_expression}'."
        )


def _create_type_to_names_dict(
    structure_expression: StructureExpression,
) -> Dict[str, List[str]]:
    # Create a dictionary with field names per type, sorted by type and then by
    # name.
    names_per_type: Dict[str, List[str]] = defaultdict(list)
    for field_match in re.findall(_REGEX_FIELD, structure_expression):
        field_name_combination, field_type = field_match[0].split(_FIELD_TYPE_POINTER)
        field_name_combination_match = re.match(
            _REGEX_FIELD_NAMES_COMBINATION, field_name_combination
        )
        if field_name_combination_match:
            field_names = field_name_combination_match.group(2).split(_SEPARATOR)
        else:
            field_names = [field_name_combination]
        names_per_type[field_type] += field_names
    return {
        field_type: sorted(names_per_type[field_type])
        for field_type in sorted(names_per_type.keys())
    }


def _type_to_names_dict_to_str(type_to_names_dict: Dict[str, List[str]]) -> str:
    # Turn the given dict into a structure expression.
    field_strings = []
    for field_type, field_names in type_to_names_dict.items():
        field_names_joined = f"{_SEPARATOR}".join(field_names)
        if len(field_names) > 1:
            field_names_joined = f"[{field_names_joined}]"
        field_strings.append(f"{field_names_joined}{_FIELD_TYPE_POINTER} {field_type}")
    return f"{_SEPARATOR}".join(field_strings)


_SEPARATOR = ","
_FIELD_TYPE_POINTER = ":"
_REGEX_SEPARATOR = rf"(\s*{_SEPARATOR}\s*)"
_REGEX_FIELD_NAME = r"(\s*[a-zA-Z]\w*\s*)"
_REGEX_FIELD_NAMES = rf"({_REGEX_FIELD_NAME}({_REGEX_SEPARATOR}{_REGEX_FIELD_NAME})+)"
_REGEX_FIELD_NAMES_COMBINATION = rf"(\s*\[{_REGEX_FIELD_NAMES}\]\s*)"
_REGEX_FIELD_LEFT = rf"({_REGEX_FIELD_NAME}|{_REGEX_FIELD_NAMES_COMBINATION})"
_REGEX_FIELD_TYPE = r"(\s*[a-zA-Z]\w*\s*)"
_REGEX_FIELD_TYPE_WILDCARD = r"(\s*\*\s*)"
_REGEX_FIELD_RIGHT = rf"({_REGEX_FIELD_TYPE}|{_REGEX_FIELD_TYPE_WILDCARD})"
_REGEX_FIELD_TYPE_POINTER = rf"(\s*{_FIELD_TYPE_POINTER}\s*)"
_REGEX_FIELD = (
    rf"(\s*{_REGEX_FIELD_LEFT}{_REGEX_FIELD_TYPE_POINTER}{_REGEX_FIELD_RIGHT}\s*)"
)
_REGEX_STRUCTURE_EXPRESSION = rf"^({_REGEX_FIELD}({_REGEX_SEPARATOR}{_REGEX_FIELD})*)$"



"""
MIT License

Copyright (c) 2022 Ramon Hagenaars

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from abc import ABC
from typing import Any, Tuple

import numpy as np

from nptyping.base_meta_classes import (
    FinalMeta,
    ImmutableMeta,
    InconstructableMeta,
    MaybeCheckableMeta,
    PrintableMeta,
    SubscriptableMeta,
)
from nptyping.error import InvalidArgumentsError
from nptyping.nptyping_type import NPTypingType
from nptyping.shape import Shape
from nptyping.shape_expression import check_shape
from nptyping.structure import Structure
from nptyping.structure_expression import check_structure, check_type_names
from nptyping.typing_ import (
    DType,
    dtype_per_name,
    name_per_dtype,
)


class NDArrayMeta(
    SubscriptableMeta,
    InconstructableMeta,
    ImmutableMeta,
    FinalMeta,
    MaybeCheckableMeta,
    PrintableMeta,
    implementation="NDArray",
):
    """
    Metaclass that is coupled to nptyping.NDArray. It contains all actual logic
    such as instance checking.
    """

    __args__: Tuple[Shape, DType]
    _parameterized: bool

    def _get_item(cls, item: Any) -> Tuple[Any, ...]:
        cls._check_item(item)
        shape, dtype = cls._get_from_tuple(item)
        return shape, dtype

    def __instancecheck__(  # pylint: disable=bad-mcs-method-argument
        self, instance: Any
    ) -> bool:
        shape, dtype = self.__args__
        dtype_is_structure = issubclass(dtype, Structure)
        structure_is_ok = dtype_is_structure and check_structure(
            instance.dtype, dtype, dtype_per_name
        )
        return (
            isinstance(instance, np.ndarray)
            and (shape is Any or check_shape(instance.shape, shape))
            and (
                dtype is Any
                or structure_is_ok
                or issubclass(instance.dtype.type, dtype)
            )
        )

    def __str__(cls) -> str:
        shape, dtype = cls.__args__
        return (
            f"{cls.__name__}[{cls._shape_expression_to_str(shape)}, "
            f"{cls._dtype_to_str(dtype)}]"
        )

    def _is_literal_like(cls, item: Any) -> bool:
        # item is a Literal or "Literal enough" (ducktyping).
        return hasattr(item, "__args__")

    def _check_item(cls, item: Any) -> None:
        # Check if the item is what we expect and raise if it is not.
        if not isinstance(item, tuple):
            raise InvalidArgumentsError(f"Unexpected argument of type {type(item)}.")
        if len(item) > 2:
            raise InvalidArgumentsError(f"Unexpected argument {item[2]}.")

    def _get_from_tuple(cls, item: Tuple[Any, ...]) -> Tuple[Shape, DType]:
        # Return the Shape Expression and DType from a tuple.
        shape = cls._get_shape(item[0])
        dtype = cls._get_dtype(item[1])
        return shape, dtype

    def _get_shape(cls, dtype_candidate: Any) -> Shape:
        if dtype_candidate is Any or dtype_candidate is Shape:
            shape = Any
        elif issubclass(dtype_candidate, Shape):
            shape = dtype_candidate
        elif cls._is_literal_like(dtype_candidate):
            shape_expression = dtype_candidate.__args__[0]
            shape = Shape[shape_expression]
        else:
            raise InvalidArgumentsError(
                f"Unexpected argument '{dtype_candidate}', expecting"
                " Shape[<ShapeExpression>]"
                " or Literal[<ShapeExpression>]"
                " or typing.Any."
            )
        return shape

    def _get_dtype(cls, dtype_candidate: Any) -> DType:
        is_dtype = isinstance(dtype_candidate, type) and issubclass(
            dtype_candidate, np.generic
        )
        if dtype_candidate is Any:
            dtype = Any
        elif is_dtype:
            dtype = dtype_candidate
        elif issubclass(dtype_candidate, Structure):
            dtype = dtype_candidate
            check_type_names(dtype, dtype_per_name)
        elif cls._is_literal_like(dtype_candidate):
            structure_expression = dtype_candidate.__args__[0]
            dtype = Structure[structure_expression]
            check_type_names(dtype, dtype_per_name)
        else:
            raise InvalidArgumentsError(
                f"Unexpected argument '{dtype_candidate}', expecting"
                " Structure[<StructureExpression>]"
                " or Literal[<StructureExpression>]"
                " or a dtype"
                " or typing.Any."
            )
        return dtype

    def _dtype_to_str(cls, dtype: Any) -> str:
        if dtype is Any:
            result = "Any"
        elif issubclass(dtype, Structure):
            result = str(dtype)
        else:
            result = name_per_dtype[dtype]
        return result

    def _shape_expression_to_str(cls, shape_expression: Any) -> str:
        return "Any" if shape_expression is Any else str(shape_expression)


class NDArray(NPTypingType, ABC, metaclass=NDArrayMeta):
    """
    An nptyping equivalent of numpy ndarray.

    ## No arguments means an NDArray with any DType and any shape.
    >>> NDArray
    NDArray[Any, Any]

    ## You can provide a DType and a Shape Expression.
    >>> from nptyping import Int32, Shape
    >>> NDArray[Shape["2, 2"], Int32]
    NDArray[Shape['2, 2'], Int]

    ## Instance checking can be done and the shape is also checked.
    >>> import numpy as np
    >>> isinstance(np.array([[1, 2], [3, 4]]), NDArray[Shape['2, 2'], Int32])
    True
    >>> isinstance(np.array([[1, 2], [3, 4], [5, 6]]), NDArray[Shape['2, 2'], Int32])
    False

    """

    __args__ = (Any, Any)



"""
MIT License

Copyright (c) 2022 Ramon Hagenaars

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from abc import ABCMeta, abstractmethod
from typing import (
    Any,
    Dict,
    Optional,
    Set,
    Tuple,
    TypeVar,
)

from nptyping.error import InvalidArgumentsError, NPTypingError

_T = TypeVar("_T")


class InconstructableMeta(ABCMeta):
    """
    Makes it impossible for a class to get instantiated.
    """

    def __call__(cls, *_: Any, **__: Any) -> None:
        raise NPTypingError(
            f"Cannot instantiate nptyping.{cls.__name__}. Did you mean to use [ ] ?"
        )


class ImmutableMeta(ABCMeta):
    """
    Makes it impossible to changes values on a class.
    """

    def __setattr__(cls, key: str, value: Any) -> None:
        if key not in ("_abc_impl", "__abstractmethods__"):
            raise NPTypingError(f"Cannot set values to nptyping.{cls.__name__}.")


class FinalMeta(ABCMeta):
    """
    Makes it impossible for classes to inherit from some class.

    An concrete inheriting meta class requires to define a name for its
    implementation. The class with this name will be the only class that is
    allowed to use that concrete meta class.
    """

    _name_per_meta_cls: Dict[type, Optional[str]] = {}

    def __init_subclass__(cls, implementation: Optional[str] = None) -> None:
        # implementation is made Optional here, to allow other meta classes to
        # inherit.
        cls._name_per_meta_cls[cls] = implementation

    def __new__(cls, name: str, *args: Any, **kwargs: Any) -> type:
        if name == cls._name_per_meta_cls[cls]:
            assert name, "cls_name not set"
            return type.__new__(cls, name, *args, **kwargs)

        raise NPTypingError(f"Cannot subclass nptyping.{cls._name_per_meta_cls[cls]}.")


class MaybeCheckableMeta(ABCMeta):
    """
    Makes instance and subclass checks raise by default.
    """

    def __instancecheck__(cls, instance: Any) -> bool:
        raise NPTypingError(
            f"Instance checking is not supported for nptyping.{cls.__name__}."
        )

    def __subclasscheck__(cls, subclass: Any) -> bool:
        raise NPTypingError(
            f"Subclass checking is not supported for nptyping.{cls.__name__}."
        )


class PrintableMeta(ABCMeta):
    """
    Ensures that a class can be printed nicely.
    """

    @abstractmethod
    def __str__(cls) -> str:
        ...  # pragma: no cover

    def __repr__(cls) -> str:
        return str(cls)


class SubscriptableMeta(ABCMeta):
    """
    Makes a class subscriptable: it accepts arguments between brackets and a
    new type is returned for every unique set of arguments.
    """

    _all_types: Dict[Tuple[type, Tuple[Any, ...]], type] = {}
    _parameterized: bool = False

    @abstractmethod
    def _get_item(cls, item: Any) -> Tuple[Any, ...]:
        ...  # pragma: no cover

    def _get_additional_values(
        cls, item: Any  # pylint: disable=unused-argument
    ) -> Dict[str, Any]:
        # This method is invoked after _get_item and right before returning
        # the result of __getitem__. It can be overridden to provide extra
        # values that are to be set as attributes on the new type.
        return {}

    def __getitem__(cls, item: Any) -> type:
        if getattr(cls, "_parameterized", False):
            raise NPTypingError(f"Type nptyping.{cls} is already parameterized.")

        args = cls._get_item(item)
        additional_values = cls._get_additional_values(item)
        assert hasattr(cls, "__args__"), "A SubscriptableMeta must have __args__."
        if args != cls.__args__:  # type: ignore[attr-defined]
            result = cls._create_type(args, additional_values)
        else:
            result = cls

        return result

    def _create_type(
        cls, args: Tuple[Any, ...], additional_values: Dict[str, Any]
    ) -> type:
        key = (cls, args)
        if key not in cls._all_types:
            cls._all_types[key] = type(
                cls.__name__,
                (cls,),
                {"__args__": args, "_parameterized": True, **additional_values},
            )
        return cls._all_types[key]


class ComparableByArgsMeta(ABCMeta):
    """
    Makes a class comparable by means of its __args__.
    """

    __args__: Tuple[Any, ...]

    def __eq__(cls, other: Any) -> bool:
        return (
            hasattr(cls, "__args__")
            and hasattr(other, "__args__")
            and cls.__args__ == other.__args__
        )

    def __hash__(cls) -> int:
        return hash(cls.__args__)


class ContainerMeta(
    InconstructableMeta,
    ImmutableMeta,
    FinalMeta,
    MaybeCheckableMeta,
    PrintableMeta,
    SubscriptableMeta,
    ComparableByArgsMeta,
    ABCMeta,
):
    """
    Base meta class for "containers" such as Shape and Structure.
    """

    _known_expressions: Set[str] = set()
    __args__: Tuple[str, ...]

    @abstractmethod
    def _validate_expression(cls, item: str) -> None:
        ...  # pragma: no cover

    @abstractmethod
    def _normalize_expression(cls, item: str) -> str:
        ...  # pragma: no cover

    def _get_item(cls, item: Any) -> Tuple[Any, ...]:
        if not isinstance(item, str):
            raise InvalidArgumentsError(
                f"Unexpected argument of type {type(item)}, expecting a string."
            )

        if item in cls._known_expressions:
            # No need to do costly validations and normalizations if it has been done
            # before.
            return (item,)

        cls._validate_expression(item)
        norm_shape_expression = cls._normalize_expression(item)
        cls._known_expressions.add(norm_shape_expression)
        return (norm_shape_expression,)

    def __subclasscheck__(cls, subclass: Any) -> bool:
        type_match = type(subclass) == type(  # pylint: disable=unidiomatic-typecheck
            cls
        )
        return type_match and (
            subclass.__args__ == cls.__args__ or not cls._parameterized
        )

    def __str__(cls) -> str:
        return f"{cls.__name__}['{cls.__args__[0]}']"

    def __eq__(cls, other: Any) -> bool:
        result = cls is other
        if not result and hasattr(cls, "__args__") and hasattr(other, "__args__"):
            normalized_args = tuple(
                cls._normalize_expression(str(arg)) for arg in other.__args__
            )
            result = cls.__args__ == normalized_args
        return result

    def __hash__(cls) -> int:
        return hash(cls.__args__)



"""
MIT License

Copyright (c) 2022 Ramon Hagenaars

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from abc import ABC
from typing import (
    Any,
    Dict,
    List,
)

from nptyping.base_meta_classes import ContainerMeta
from nptyping.nptyping_type import NPTypingType
from nptyping.structure_expression import (
    create_name_to_type_dict,
    normalize_structure_expression,
    validate_structure_expression,
)


class StructureMeta(ContainerMeta, implementation="Structure"):
    """
    Metaclass that is coupled to nptyping.Structure.
    """

    __args__ = tuple()

    def _validate_expression(cls, item: str) -> None:
        validate_structure_expression(item)

    def _normalize_expression(cls, item: str) -> str:
        return normalize_structure_expression(item)

    def _get_additional_values(cls, item: Any) -> Dict[str, Any]:
        return {"_type_per_name": create_name_to_type_dict(item)}


class Structure(NPTypingType, ABC, metaclass=StructureMeta):
    """
    A container for structure expressions that describe the structured dtype of
    an array.

    Simple example:

    >>> Structure["x: Float, y: Float"]
    Structure['[x, y]: Float']

    """

    _type_per_name = {}

    @classmethod
    def get_types(cls) -> List[str]:
        """
        Return a list of all types (strings) in this Structure.
        :return: a list of all types in this Structure.
        """
        return list(set(cls._type_per_name.values()))

    @classmethod
    def get_names(cls) -> List[str]:
        """
        Return a list of all names in this Structure.
        :return: a list of all names in this Structure.
        """
        return list(cls._type_per_name.keys())

    @classmethod
    def get_type(cls, name: str) -> str:
        """
        Get the type (str) that corresponds to the given name. For example for
        Structure["x: Float"], get_type("x") would give "Float".
        :param name: the name of which the type is to be returned.
        :return: the type as a string that corresponds to that name.
        """
        return cls._type_per_name[name]



"""
MIT License

Copyright (c) 2022 Ramon Hagenaars

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

try:
    from typing import (  # type: ignore[attr-defined,misc] # pylint: disable=unused-import
        Literal,
        TypeAlias,
        TypeGuard,
        final,
    )
except ImportError:  # pragma: no cover
    from typing_extensions import (  # type: ignore[attr-defined,misc]
        Literal,
        TypeAlias,
        TypeGuard,
        final,
    )

from typing import Tuple, Union

import numpy as np

ShapeExpression: TypeAlias = str
StructureExpression: TypeAlias = str
DType: TypeAlias = Union[np.generic, StructureExpression]
ShapeTuple: TypeAlias = Tuple[int, ...]

Number = np.number
Bool = np.bool_
Bool8 = np.bool8
Object = np.object_
Object0 = np.object0
Datetime64 = np.datetime64
Integer = np.integer
SignedInteger = np.signedinteger
Int8 = np.int8
Int16 = np.int16
Int32 = np.int32
Int64 = np.int64
Byte = np.byte
Short = np.short
IntC = np.intc
IntP = np.intp
Int0 = np.int0
Int = np.int_
LongLong = np.longlong
Timedelta64 = np.timedelta64
UnsignedInteger = np.unsignedinteger
UInt8 = np.uint8
UInt16 = np.uint16
UInt32 = np.uint32
UInt64 = np.uint64
UByte = np.ubyte
UShort = np.ushort
UIntC = np.uintc
UIntP = np.uintp
UInt0 = np.uint0
UInt = np.uint
ULongLong = np.ulonglong
Inexact = np.inexact
Floating = np.floating
Float16 = np.float16
Float32 = np.float32
Float64 = np.float64
Half = np.half
Single = np.single
Double = np.double
Float = np.float_
LongDouble = np.longdouble
LongFloat = np.longfloat
ComplexFloating = np.complexfloating
Complex64 = np.complex64
Complex128 = np.complex128
CSingle = np.csingle
SingleComplex = np.singlecomplex
CDouble = np.cdouble
Complex = np.complex_
CFloat = np.cfloat
CLongDouble = np.clongdouble
CLongFloat = np.clongfloat
LongComplex = np.longcomplex
Flexible = np.flexible
Void = np.void
Void0 = np.void0
Character = np.character
Bytes = np.bytes_
Str = np.str_
String = np.string_
Bytes0 = np.bytes0
Unicode = np.unicode_
Str0 = np.str0

dtypes = [
    (Number, "Number"),
    (Bool, "Bool"),
    (Bool8, "Bool8"),
    (Object, "Object"),
    (Object0, "Object0"),
    (Datetime64, "Datetime64"),
    (Integer, "Integer"),
    (SignedInteger, "SignedInteger"),
    (Int8, "Int8"),
    (Int16, "Int16"),
    (Int32, "Int32"),
    (Int64, "Int64"),
    (Byte, "Byte"),
    (Short, "Short"),
    (IntC, "IntC"),
    (IntP, "IntP"),
    (Int0, "Int0"),
    (Int, "Int"),
    (LongLong, "LongLong"),
    (Timedelta64, "Timedelta64"),
    (UnsignedInteger, "UnsignedInteger"),
    (UInt8, "UInt8"),
    (UInt16, "UInt16"),
    (UInt32, "UInt32"),
    (UInt64, "UInt64"),
    (UByte, "UByte"),
    (UShort, "UShort"),
    (UIntC, "UIntC"),
    (UIntP, "UIntP"),
    (UInt0, "UInt0"),
    (UInt, "UInt"),
    (ULongLong, "ULongLong"),
    (Inexact, "Inexact"),
    (Floating, "Floating"),
    (Float16, "Float16"),
    (Float32, "Float32"),
    (Float64, "Float64"),
    (Half, "Half"),
    (Single, "Single"),
    (Double, "Double"),
    (Float, "Float"),
    (LongDouble, "LongDouble"),
    (LongFloat, "LongFloat"),
    (ComplexFloating, "ComplexFloating"),
    (Complex64, "Complex64"),
    (Complex128, "Complex128"),
    (CSingle, "CSingle"),
    (SingleComplex, "SingleComplex"),
    (CDouble, "CDouble"),
    (Complex, "Complex"),
    (CFloat, "CFloat"),
    (CLongDouble, "CLongDouble"),
    (CLongFloat, "CLongFloat"),
    (LongComplex, "LongComplex"),
    (Flexible, "Flexible"),
    (Void, "Void"),
    (Void0, "Void0"),
    (Character, "Character"),
    (Bytes, "Bytes"),
    (String, "String"),
    (Str, "Str"),
    (Bytes0, "Bytes0"),
    (Unicode, "Unicode"),
    (Str0, "Str0"),
]

name_per_dtype = dict(dtypes)
dtype_per_name = {name: dtype for dtype, name in dtypes}



import sys
from glob import glob
from pathlib import Path

from setuptools import find_packages, setup

project_slug = "nptyping"
here = Path(__file__).parent.absolute()


def _get_dependencies(dependency_file):
    with open(here / "dependencies" / dependency_file, mode="r", encoding="utf-8") as f:
        return f.read().strip().split("\n")


# Read meta info from package_info.py.
package_info = {}
with open(here / project_slug / "package_info.py", mode="r", encoding="utf-8") as f:
    exec(f.read(), package_info)
supp_versions = package_info["__python_versions__"]

# The README.md provides the long description text.
with open("README.md", mode="r", encoding="utf-8") as f:
    long_description = f.read()

# Check the current version against the supported versions: older versions are not supported.
u_major = sys.version_info.major
u_minor = sys.version_info.minor
versions_as_ints = [[int(v) for v in version.split(".")] for version in supp_versions]
version_unsupported = not [
    1 for major, minor in versions_as_ints if u_major == major and u_minor >= minor
]
if version_unsupported:
    supported_versions_str = ", ".join(version for version in supp_versions)
    raise Exception(
        f"Unsupported Python version: {sys.version}. Supported versions: {supported_versions_str}"
    )


extras = {
    "build": _get_dependencies("build-requirements.txt"),
    "qa": _get_dependencies("qa-requirements.txt"),
}
# Complete: all extras for end users, excluding dev dependencies.
extras["complete"] = [
    req for key, reqs in extras.items() for req in reqs if key not in ("build", "qa")
]
# Dev: all extras for developers, including build and qa dependencies.
extras["dev"] = [req for key, reqs in extras.items() for req in reqs if key]

pyi_files = [
    Path(pyi_file).name for pyi_file in glob("nptyping/**/*.pyi", recursive=True)
]

setup(
    name=package_info["__title__"],
    version=package_info["__version__"],
    author=package_info["__author__"],
    author_email=package_info["__author_email__"],
    description=package_info["__description__"],
    url=package_info["__url__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    license=package_info["__license__"],
    package_data={
        "nptyping": pyi_files + ["py.typed"],
    },
    packages=find_packages(
        exclude=("tests", "tests.*", "test_resources", "test_resources.*")
    ),
    install_requires=_get_dependencies("requirements.txt"),
    extras_require=extras,
    python_requires=f">={supp_versions[0]}",
    test_suite="tests",
    zip_safe=False,
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        *[f"Programming Language :: Python :: {version}" for version in supp_versions],
    ],
)
